@@ -1,8 +1,9 @@
 from numpy.testing import *
-from environment import Settings
+from .environment import Settings
 import os
-
+import pytest
 import numpy
+numpy.random.seed(0)
 
 import algopy
 from algopy.tracer.tracer import *
@@ -145,6 +146,48 @@ class Test_Function_on_UTPM(TestCase):
         assert_array_almost_equal(x1.xbar.data, x2.xbar.data)
 
 
+    def test_prod(self):
+        x = numpy.random.random(4)
+
+        def f(x):
+
+            return algopy.prod(x)
+
+        cg = algopy.CGraph()
+        fx = algopy.Function(x)
+        fd = f(fx)
+        cg.independentFunctionList = [fx]
+        cg.dependentFunctionList = [fd]
+
+        grad = cg.gradient(x)
+
+
+        # test forward mode
+        ux = algopy.UTPM.init_jacobian(x)
+        uy = algopy.UTPM.prod(ux)
+        jac = algopy.UTPM.extract_jacobian(uy).reshape(x.shape)
+
+        assert_almost_equal(grad, jac)
+
+    def test_fft(self):
+
+        def eval_f2(x):
+            y = algopy.fft.fft(x, axis=0)
+            z = algopy.fft.ifft(y, axis=0)
+            return algopy.real(z)
+
+        signal = numpy.array([1,2,3,4,5.])
+
+        cg = algopy.CGraph()
+        x = algopy.Function(signal)
+        y = eval_f2(x)
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+
+        algopy_jacobian = cg.jacobian(signal)
+
+        assert_almost_equal(algopy_jacobian, numpy.eye(signal.size))
 
 
     def test_sqrt_in_norm_computation(self):
@@ -383,6 +426,29 @@ class Test_CGgraph_on_UTPM(TestCase):
 
         assert_array_almost_equal(x.data,UTPM.dot(Q,R).data)
 
+    def test_lu(self):
+        def f(x):
+
+            W, L, U = algopy.lu(x)
+
+            return algopy.sum(algopy.diag(U))
+
+        # reverse mode
+        x = numpy.random.random((2,2))
+        cg = algopy.CGraph()
+        fx = algopy.Function(x)
+        fd = f(fx)
+        cg.independentFunctionList = [fx]
+        cg.dependentFunctionList = [fd]
+
+        grad = cg.gradient(x)
+
+        # test forward mode
+        ux = algopy.UTPM.init_jacobian(x)
+        uy = f(ux)
+        jac = algopy.UTPM.extract_jacobian(uy).reshape(x.shape)
+        assert_almost_equal(jac, grad)
+
     def test_pullback_symvec_vecsym(self):
         (D,P,N) = 2,1,6
         cg = CGraph()
@@ -1147,6 +1213,96 @@ class Test_CGgraph_on_UTPM(TestCase):
 
         assert_almost_equal(numpy.trace(numpy.dot(Abar.T, Adot)), numpy.trace( numpy.dot(Lbar.T, Ldot) + numpy.dot(Qbar.T, Qdot)))
 
+    def test_complex1(self):
+
+        import algopy
+        import numpy as np
+
+        def f(x, A, module):
+              y = module.dot(A, x)
+              return module.real(module.sum(y))
+
+        size = 4
+        Ar = np.random.random((size, size))
+        Ai = np.random.random((size, size))
+        Ac = Ar +1j*Ai
+        A  = Ac
+        x  = np.random.random((size,))
+
+        cg = algopy.CGraph()
+        xf = algopy.Function(x)
+        sf = f(xf, A, algopy)
+        cg.trace_off()
+        assert_array_almost_equal(sf.x, f(x , A, np))
+
+        cg.independentFunctionList = [xf]
+        cg.dependentFunctionList = [sf]
+        gf = cg.gradient(x)
+        ganalytic = np.real(np.sum(A, axis=0))
+
+        assert_array_almost_equal(gf, ganalytic)
+
+
+    def test_complex2(self):
+
+        import algopy
+        import numpy as np
+
+        def f(x, A, module):
+              y = module.dot(A, x)
+              return module.imag(module.sum(y))
+
+        size = 4
+        Ar = np.random.random((size, size))
+        Ai = np.random.random((size, size))
+        Ac = Ar +1j*Ai
+        A  = Ac
+        x  = np.random.random((size,))
+
+        cg = algopy.CGraph()
+        xf = algopy.Function(x)
+        sf = f(xf, A, algopy)
+        cg.trace_off()
+
+        assert_array_almost_equal(sf.x, f(x , A, np))
+
+
+        cg.independentFunctionList = [xf]
+        cg.dependentFunctionList = [sf]
+        gf = cg.gradient(x)
+        ganalytic = np.imag(np.sum(A, axis=0))
+        assert_array_almost_equal(gf, ganalytic)
+
+
+    def test_complex3(self):
+
+        import algopy
+        import numpy as np
+
+        def f(x, A, module):
+              y = module.dot(A, x)
+              return module.sum(y)
+
+        size = 4
+        Ar = np.random.random((size, size))
+        Ai = np.random.random((size, size))
+        Ac = Ar +1j*Ai
+        A  = Ac
+        x  = np.random.random((size,)) + 0j
+
+        cg = algopy.CGraph()
+        xf = algopy.Function(x)
+        sf = f(xf, A, algopy)
+        cg.trace_off()
+
+        assert_array_almost_equal(sf.x, f(x , A, np))
+
+        cg.independentFunctionList = [xf]
+        cg.dependentFunctionList = [sf]
+        gf = cg.gradient(x)
+        ganalytic = np.sum(A, axis=0)
+
+        assert_array_almost_equal(gf, ganalytic)
 
 
 
@@ -1183,7 +1339,7 @@ class Test_CGgraph_on_UTPM(TestCase):
             return numpy.array([ y[1]/y[2],
                                  (y[0] + 2*y[1])/y[2],
                                  -(y[0] + y[1]) * y[1]/(y[2]*y[2])
-                                 ])
+                                 ], dtype=object)
 
 
         # check correctness of the push forward
@@ -1239,7 +1395,7 @@ class Test_CGgraph_on_UTPM(TestCase):
         assert_array_almost_equal(const1.data[0,:], const2.data[0,:])
 
 
-    @decorators.skipif(mpmath is None)
+    @pytest.mark.skipif(mpmath is None, reason='requires mpmath to run')
     def test_dpm_hyp1f1(self):
         """
         compute y = dpm_hyp1f1(1., 2., x**2 + 3.)
@@ -1339,72 +1495,72 @@ class Test_CGgraph_on_UTPM(TestCase):
         assert_array_almost_equal(result3, result1)
 
 
-    @decorators.skipif(mpmath is None)
-    def test_dpm_hyp2f0(self):
-        """
-        compute y = dpm_hyp2f0(0.5, 1.0, 0.1 * x**2 + 0.03)
-        """
+    # @decorators.skipif(mpmath is None)
+    # def test_dpm_hyp2f0(self):
+    #     """
+    #     compute y = dpm_hyp2f0(0.5, 1.0, 0.1 * x**2 + 0.03)
+    #     """
+    #
+    #     def f(x):
+    #         # use smaller offset to ameliorate convergence issues
+    #         v1 = 0.1 * x**2 + 0.03
+    #         y = algopy.special.dpm_hyp2f0(0.5, 1.0, v1)
+    #         return y
+    #
+    #     # use CGraph
+    #
+    #     cg = CGraph()
+    #     x = Function(numpy.array([1.]))
+    #     y = f(x)
+    #
+    #     cg.independentFunctionList = [x]
+    #     cg.dependentFunctionList = [y]
+    #
+    #     result1 = cg.jac_vec(numpy.array([2.]), numpy.array([1.]))
+    #     result2 = cg.jacobian(numpy.array([2.]))[0]
+    #
+    #     # use UTPM
+    #
+    #     x = UTPM.init_jacobian(numpy.array([2.]))
+    #     y = f(x)
+    #     result3 = UTPM.extract_jacobian(y)[0]
+    #
+    #     assert_array_almost_equal(result1, result2)
+    #     assert_array_almost_equal(result2, result3)
+    #     assert_array_almost_equal(result3, result1)
 
-        def f(x):
-            # use smaller offset to ameliorate convergence issues
-            v1 = 0.1 * x**2 + 0.03
-            y = algopy.special.dpm_hyp2f0(0.5, 1.0, v1)
-            return y
-
-        # use CGraph
-
-        cg = CGraph()
-        x = Function(numpy.array([1.]))
-        y = f(x)
-
-        cg.independentFunctionList = [x]
-        cg.dependentFunctionList = [y]
-
-        result1 = cg.jac_vec(numpy.array([2.]), numpy.array([1.]))
-        result2 = cg.jacobian(numpy.array([2.]))[0]
-
-        # use UTPM
-
-        x = UTPM.init_jacobian(numpy.array([2.]))
-        y = f(x)
-        result3 = UTPM.extract_jacobian(y)[0]
-
-        assert_array_almost_equal(result1, result2)
-        assert_array_almost_equal(result2, result3)
-        assert_array_almost_equal(result3, result1)
-
-    def test_hyp2f0(self):
-        """
-        compute y = hyp2f0(0.5, 1.0, 0.1 * x**2 + 0.03)
-        """
-
-        def f(x):
-            # use smaller offset to ameliorate convergence issues
-            v1 = 0.1 * x**2 + 0.03
-            y = algopy.special.hyp2f0(0.5, 1.0, v1)
-            return y
-
-        # use CGraph
-
-        cg = CGraph()
-        x = Function(numpy.array([1.]))
-        y = f(x)
-
-        cg.independentFunctionList = [x]
-        cg.dependentFunctionList = [y]
-
-        result1 = cg.jac_vec(numpy.array([2.]), numpy.array([1.]))
-        result2 = cg.jacobian(numpy.array([2.]))[0]
-
-        # use UTPM
-
-        x = UTPM.init_jacobian(numpy.array([2.]))
-        y = f(x)
-        result3 = UTPM.extract_jacobian(y)[0]
-
-        assert_array_almost_equal(result1, result2)
-        assert_array_almost_equal(result2, result3)
-        assert_array_almost_equal(result3, result1)
+    # def test_hyp2f0(self):
+    #     """
+    #     compute y = hyp2f0(0.5, 1.0, 0.1 * x**2 + 0.03)
+    #     """
+    #
+    #     def f(x):
+    #         # use smaller offset to ameliorate convergence issues
+    #         v1 = 0.1 * x**2 + 0.03
+    #         y = algopy.special.hyp2f0(0.5, 1.0, v1)
+    #         return y
+    #
+    #     # use CGraph
+    #
+    #     cg = CGraph()
+    #     x = Function(numpy.array([1.]))
+    #     y = f(x)
+    #
+    #     cg.independentFunctionList = [x]
+    #     cg.dependentFunctionList = [y]
+    #
+    #     result1 = cg.jac_vec(numpy.array([2.]), numpy.array([1.]))
+    #     result2 = cg.jacobian(numpy.array([2.]))[0]
+    #
+    #     # use UTPM
+    #
+    #     x = UTPM.init_jacobian(numpy.array([2.]))
+    #     y = f(x)
+    #     result3 = UTPM.extract_jacobian(y)[0]
+    #
+    #     assert_array_almost_equal(result1, result2)
+    #     assert_array_almost_equal(result2, result3)
+    #     assert_array_almost_equal(result3, result1)
 
 
     def test_hyp0f1(self):
@@ -1990,7 +2146,71 @@ class Test_CGgraph_on_UTPM(TestCase):
         assert_array_almost_equal(result2, result3)
         assert_array_almost_equal(result3, result1)
 
+    def test_logdet(self):
+        x = numpy.random.random((3,3))
+        x = x.dot(x.T)
 
+        def f(x):
+
+            return algopy.logdet(x)
+
+        # reverse mode
+        cg = algopy.CGraph()
+        fx = algopy.Function(x)
+        fd = f(fx)
+        cg.independentFunctionList = [fx]
+        cg.dependentFunctionList = [fd]
+        grad = cg.gradient(x)
+
+
+        # forward mode
+        ux = algopy.UTPM.init_jacobian(x)
+        uy = algopy.UTPM.logdet(ux)
+        jac = algopy.UTPM.extract_jacobian(uy).reshape(x.shape)
+
+        # compare forward to reverse mode
+        assert_almost_equal(jac, grad)
+
+    def test_compare_det_and_logdet(self):
+
+        def f(x):
+            return algopy.log(algopy.det(x))
+
+        def g(x):
+            return algopy.logdet(x)
+
+
+        x = numpy.random.random((2,2))
+        x = x.dot(x.T) # make sure that det >= 0
+
+        cg = algopy.CGraph()
+        fx = algopy.Function(x)
+        fd = f(fx)
+        cg.independentFunctionList = [fx]
+        cg.dependentFunctionList = [fd]
+
+        cg2 = algopy.CGraph()
+        fx = algopy.Function(x)
+        fd = g(fx)
+        cg2.independentFunctionList = [fx]
+        cg2.dependentFunctionList = [fd]
+
+        grad  = cg.gradient(x)
+        grad2 = cg2.gradient(x)
+        assert_almost_equal(grad, grad2)
+
+        # forward mode
+        ux = algopy.UTPM.init_jacobian(x)
+        uy = f(ux)
+        jac = algopy.UTPM.extract_jacobian(uy).reshape(x.shape)
+
+        # forward mode
+        ux = algopy.UTPM.init_jacobian(x)
+        uy = g(ux)
+        jac2 = algopy.UTPM.extract_jacobian(uy).reshape(x.shape)
+
+        assert_almost_equal(jac, jac2)
+        assert_almost_equal(jac, grad)
 
     def test_more_complicated_ODOE_objective_function(self):
         """
@@ -2358,5 +2578,3 @@ class Test_CGraph_Plotting(TestCase):
 
 
 
-if __name__ == "__main__":
-    run_module_suite()
